@@ -7,17 +7,21 @@ from orca_python.main import (
     ArrayResult,
     ValueResult,
     StructResult,
+    MetadataField,
     ExecutionParams,
+    lookback,
 )
 
 __all__ = [
     "Processor",
     "EmitWindow",
     "Window",
+    "MetadataField",
     "WindowType",
     "StructResult",
     "ValueResult",
     "ArrayResult",
     "NoneResult",
     "ExecutionParams",
+    "lookback",
 ]
